/*
	* Copyright (C) 2010-2021 Sebastiano Vigna
	*
	* Licensed under the Apache License, Version 2.0 (the "License");
	* you may not use this file except in compliance with the License.
	* You may obtain a copy of the License at
	*
	*     http://www.apache.org/licenses/LICENSE-2.0
	*
	* Unless required by applicable law or agreed to in writing, software
	* distributed under the License is distributed on an "AS IS" BASIS,
	* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	* See the License for the specific language governing permissions and
	* limitations under the License.
	*/
package it.unimi.dsi.fastutil.floats;
import static it.unimi.dsi.fastutil.BigArrays.ensureOffsetLength;
import static it.unimi.dsi.fastutil.BigArrays.length;
import it.unimi.dsi.fastutil.BigArrays;
import java.util.Iterator;
import java.util.Collection;
import java.util.NoSuchElementException;
import it.unimi.dsi.fastutil.BigList;
import it.unimi.dsi.fastutil.BigListIterator;
/**  An abstract class providing basic methods for big lists implementing a type-specific big list interface.
	*
	* <p>Most of the methods in this class are optimized with the assumption that the List will have
	* {@link java.util.RandomAccess have constant-time random access}. If this is not the case, you
	* should probably <em>at least</em> override {@link #listIterator(long)} and the {@code xAll()} methods
	* (such as {@link #addAll}) with a more appropriate iteration scheme. Note the {@link #subList(long, long)}
	* method is cognizant of random-access or not, so that need not be reimplemented.
	*/
public abstract class AbstractFloatBigList extends AbstractFloatCollection implements FloatBigList , FloatStack {
	protected AbstractFloatBigList() {}
	/** Ensures that the given index is nonnegative and not greater than this big-list size.
	 *
	 * @param index an index.
	 * @throws IndexOutOfBoundsException if the given index is negative or greater than this big-list size.
	 */
	protected void ensureIndex(final long index) {
	 if (index < 0) throw new IndexOutOfBoundsException("Index (" + index + ") is negative");
	 if (index > size64()) throw new IndexOutOfBoundsException("Index (" + index + ") is greater than list size (" + (size64()) + ")");
	}
	/** Ensures that the given index is nonnegative and smaller than this big-list size.
	 *
	 * @param index an index.
	 * @throws IndexOutOfBoundsException if the given index is negative or not smaller than this big-list size.
	 */
	protected void ensureRestrictedIndex(final long index) {
	 if (index < 0) throw new IndexOutOfBoundsException("Index (" + index + ") is negative");
	 if (index >= size64()) throw new IndexOutOfBoundsException("Index (" + index + ") is greater than or equal to list size (" + (size64()) + ")");
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation always throws an {@link UnsupportedOperationException}.
	 */
	@Override
	public void add(final long index, final float k) {
	 throw new UnsupportedOperationException();
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the type-specific version of {@link BigList#add(long, Object)}.
	 */
	@Override
	public boolean add(final float k) {
	 add(size64(), k);
	 return true;
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation always throws an {@link UnsupportedOperationException}.
	 */
	@Override
	public float removeFloat(long i) {
	 throw new UnsupportedOperationException();
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation always throws an {@link UnsupportedOperationException}.
	 */
	@Override
	public float set(final long index, final float k) {
	 throw new UnsupportedOperationException();
	}
	/** Adds all of the elements in the specified collection to this list (optional operation). */
	@Override
	public boolean addAll(long index, final Collection<? extends Float> c) {
	 ensureIndex(index);
	 final Iterator<? extends Float> i = c.iterator();
	 final boolean retVal = i.hasNext();
	 while(i.hasNext()) add(index++, i.next());
	 return retVal;
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the type-specific version of {@link BigList#addAll(long, Collection)}.
	 */
	@Override
	public boolean addAll(final Collection<? extends Float> c) {
	 return addAll(size64(), c);
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to {@link #listIterator()}.
	 */
	@Override
	public FloatBigListIterator iterator() {
	 return listIterator();
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to {@link BigList#listIterator(long) listIterator(0)}.
	 */
	@Override
	public FloatBigListIterator listIterator() {
	 return listIterator(0L);
	}
	/** {@inheritDoc}
	 * @implSpec This implementation is based on the random-access methods. */
	@Override
	public FloatBigListIterator listIterator(final long index) {
	 ensureIndex(index);
	 return new FloatBigListIterators.AbstractIndexBasedBigListIterator (0, index) {
	  @Override
	  protected final float get(long i) { return AbstractFloatBigList.this.getFloat(i); }
	  @Override
	  protected final void add(long i, float k) { AbstractFloatBigList.this.add(i, k); }
	  @Override
	  protected final void set(long i, float k) { AbstractFloatBigList.this.set(i, k); }
	  @Override
	  protected final void remove(long i) { AbstractFloatBigList.this.removeFloat(i); }
	  @Override
	  protected final long getMaxPos() { return AbstractFloatBigList.this.size64(); }
	 };
	}
	static final class IndexBasedSpliterator extends FloatBigSpliterators.LateBindingSizeIndexBasedSpliterator {
	 final FloatBigList l;
	 IndexBasedSpliterator(FloatBigList l, long pos) {
	  super(pos);
	  this.l = l;
	 }
	 IndexBasedSpliterator(FloatBigList l, long pos, long maxPos) {
	  super(pos, maxPos);
	  this.l = l;
	 }
	 @Override
	 protected final long getMaxPosFromBackingStore() { return l.size64(); }
	 @Override
	 protected final float get(long i) { return l.getFloat(i); }
	 @Override
	 protected final IndexBasedSpliterator makeForSplit(long pos, long maxPos) {
	  return new IndexBasedSpliterator (l, pos, maxPos);
	 }
	}
	@Override
	public it.unimi.dsi.fastutil.doubles.DoubleSpliterator doubleSpliterator() {
	 if (this instanceof java.util.RandomAccess) {
	  return FloatSpliterators.widen(spliterator());
	 } else {
	  return super.doubleSpliterator();
	 }
	}
	/** Returns true if this list contains the specified element.
	 * @implSpec This implementation delegates to {@code indexOf()}.
	 * @see BigList#contains(Object)
	 */
	@Override
	public boolean contains(final float k) { return indexOf(k) >= 0; }
	@Override
	public long indexOf(final float k) {
	 final FloatBigListIterator i = listIterator();
	 float e;
	 while(i.hasNext()) {
	  e = i.nextFloat();
	  if (( Float.floatToIntBits(k) == Float.floatToIntBits(e) )) return i.previousIndex();
	 }
	 return -1;
	}
	@Override
	public long lastIndexOf(final float k) {
	 FloatBigListIterator i = listIterator(size64());
	 float e;
	 while(i.hasPrevious()) {
	  e = i.previousFloat();
	  if (( Float.floatToIntBits(k) == Float.floatToIntBits(e) )) return i.nextIndex();
	 }
	 return -1;
	}
	@Override
	public void size(final long size) {
	 long i = size64();
	 if (size > i) while(i++ < size) add((0));
	 else while(i-- != size) remove(i);
	}
	@Override
	public FloatBigList subList(final long from, final long to) {
	 ensureIndex(from);
	 ensureIndex(to);
	 if (from > to) throw new IndexOutOfBoundsException("Start index (" + from + ") is greater than end index (" + to + ")");
	 return this instanceof java.util.RandomAccess ? new FloatRandomAccessSubList (this, from, to) : new FloatSubList (this, from, to);
	}
	/** {@inheritDoc}
	 *
	 * @implSpec If this list is {@link java.util.RandomAccess}, will iterate using a for
	 * loop and the type-specific {@link java.util.List#get(int)} method. Otherwise it will fallback
	 * to using the iterator based loop implementation from the superinterface.
	 */
	 @Override
	 public void forEach(final FloatConsumer action) {
	  if (this instanceof java.util.RandomAccess) {
	   for (long i = 0, max = size64(); i < max; ++i) {
	    action.accept(getFloat(i));
	  }
	 } else {
	  super.forEach(action);
	 }
	}
	/** {@inheritDoc}
	 *
	 * <p>This is a trivial iterator-based implementation. It is expected that
	 * implementations will override this method with a more optimized version.
	 */
	@Override
	public void removeElements(final long from, final long to) {
	 ensureIndex(to);
	 FloatBigListIterator i = listIterator(from);
	 long n = to - from;
	 if (n < 0) throw new IllegalArgumentException("Start index (" + from + ") is greater than end index (" + to + ")");
	 while(n-- != 0) {
	  i.nextFloat();
	  i.remove();
	 }
	}
	/** {@inheritDoc}
	 *
	 * <p>This is a trivial iterator-based implementation. It is expected that
	 * implementations will override this method with a more optimized version.
	 */
	@Override
	public void addElements(long index, final float a[][], long offset, long length) {
	 ensureIndex(index);
	 ensureOffsetLength(a, offset, length);
	 if (this instanceof java.util.RandomAccess) {
	  while(length-- != 0) add(index++, BigArrays.get(a, offset++));
	 } else {
	  FloatBigListIterator iter = listIterator(index);
	  while(length-- != 0) iter.add(BigArrays.get(a, offset++));
	 }
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the analogous method for big-array fragments.
	 */
	@Override
	public void addElements(final long index, final float a[][]) {
	 addElements(index, a, 0, length(a));
	}
	/** {@inheritDoc}
	 *
	 * <p>This is a trivial iterator-based implementation. It is expected that
	 * implementations will override this method with a more optimized version.
	 */
	@Override
	public void getElements(final long from, final float a[][], long offset, long length) {
	 ensureIndex(from);
	 ensureOffsetLength(a, offset, length);
	 if (from + length > size64()) throw new IndexOutOfBoundsException("End index (" + (from + length) + ") is greater than list size (" + size64() + ")");
	 if (this instanceof java.util.RandomAccess) {
	  long current = from;
	  while(length-- != 0) BigArrays.set(a, offset++, getFloat(current++));
	 } else {
	  FloatBigListIterator i = listIterator(from);
	  while(length-- != 0) BigArrays.set(a, offset++, i.nextFloat());
	 }
	}
	@Override
	public void setElements(long index, float a[][], long offset, long length) {
	 ensureIndex(index);
	 ensureOffsetLength(a, offset, length);
	 if (index + length > size64()) throw new IndexOutOfBoundsException("End index (" + (index + length) + ") is greater than list size (" + size64() + ")");
	 if (this instanceof java.util.RandomAccess) {
	  for (long i = 0; i < length; ++i) {
	   set(i + index, BigArrays.get(a, i + offset));
	  }
	 } else {
	  FloatBigListIterator iter = listIterator(index);
	  long i = 0;
	  while (i < length) {
	   iter.nextFloat();
	   iter.set(BigArrays.get(a, offset + i++));
	  }
	 }
	}
	/** {@inheritDoc}
	 * @implSpec This implementation delegates to {@link #removeElements(long, long)}.*/
	@Override
	public void clear() {
	 removeElements(0, size64());
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to {@link #size64()}.
	 * @deprecated Please use {@link #size64()} instead. */
	@Override
	@Deprecated
	public int size() {
	 return (int)Math.min(Integer.MAX_VALUE, size64());
	}
	/** Returns the hash code for this big list, which is identical to {@link java.util.List#hashCode()}.
	 *
	 * @return the hash code for this big list.
	 */
	@Override
	public int hashCode() {
	 FloatIterator i = iterator();
	 int h = 1;
	 long s = size64();
	 while (s-- != 0) {
	  float k = i.nextFloat();
	  h = 31 * h + it.unimi.dsi.fastutil.HashCommon.float2int(k);
	 }
	 return h;
	}
	@Override
	public boolean equals(final Object o) {
	 if (o == this) return true;
	 if (! (o instanceof BigList)) return false;
	 final BigList<?> l = (BigList<?>)o;
	 long s = size64();
	 if (s != l.size64()) return false;
	 if (l instanceof FloatBigList) {
	  final FloatBigListIterator i1 = listIterator(), i2 = ((FloatBigList )l).listIterator();
	  while(s-- != 0) if (i1.nextFloat() != i2.nextFloat()) return false;
	  return true;
	 }
	 final BigListIterator<?> i1 = listIterator(), i2 = l.listIterator();
	 while(s-- != 0) if (! java.util.Objects.equals(i1.next(), i2.next())) return false;
	 return true;
	}
	/** Compares this big list to another object. If the
	 * argument is a {@link BigList}, this method performs a lexicographical comparison; otherwise,
	 * it throws a {@code ClassCastException}.
	 *
	 * @param l a big list.
	 * @return if the argument is a {@link BigList}, a negative integer,
	 * zero, or a positive integer as this list is lexicographically less than, equal
	 * to, or greater than the argument.
	 * @throws ClassCastException if the argument is not a big list.
	 */

	@Override
	public int compareTo(final BigList<? extends Float> l) {
	 if (l == this) return 0;
	 if (l instanceof FloatBigList) {
	  final FloatBigListIterator i1 = listIterator(), i2 = ((FloatBigList )l).listIterator();
	  int r;
	  float e1, e2;
	  while(i1.hasNext() && i2.hasNext()) {
	   e1 = i1.nextFloat();
	   e2 = i2.nextFloat();
	   if ((r = ( Float.compare((e1),(e2)) )) != 0) return r;
	  }
	  return i2.hasNext() ? -1 : (i1.hasNext() ? 1 : 0);
	 }
	 BigListIterator<? extends Float> i1 = listIterator(), i2 = l.listIterator();
	 int r;
	 while(i1.hasNext() && i2.hasNext()) {
	  if ((r = ((Comparable<? super Float>)i1.next()).compareTo(i2.next())) != 0) return r;
	 }
	 return i2.hasNext() ? -1 : (i1.hasNext() ? 1 : 0);
	}
	@Override
	public void push(float o) {
	 add(o);
	}
	@Override
	public float popFloat() {
	 if (isEmpty()) throw new NoSuchElementException();
	 return removeFloat(size64() - 1);
	}
	@Override
	public float topFloat() {
	 if (isEmpty()) throw new NoSuchElementException();
	 return getFloat(size64() - 1);
	}
	@Override
	public float peekFloat(int i) {
	 return getFloat(size64() - 1 - i);
	}
	/** Removes a single instance of the specified element from this collection, if it is present (optional operation).
	 * @implSpec This implementation delegates to {@code indexOf()}.
	 * @see BigList#remove(Object)
	 */
	@Override
	public boolean rem(float k) {
	 long index = indexOf(k);
	 if (index == -1) return false;
	 removeFloat(index);
	 return true;
	}
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the type-specific version of {@link #addAll(long, Collection)}.
	 */
	@Override
	public boolean addAll(final long index, final FloatCollection c) { return addAll(index, (Collection<? extends Float>)c); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the type-specific version of {@link #addAll(long, Collection)}.
	 */
	@Override
	public boolean addAll(final FloatCollection c) { return addAll(size64(), c); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public void add(final long index, final Float ok) { add(index, ok.floatValue()); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public Float set(final long index, final Float ok) { return Float.valueOf(set(index, ok.floatValue())); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public Float get(final long index) { return Float.valueOf(getFloat(index)); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public long indexOf(final Object ok) { return indexOf(((Float)(ok)).floatValue()); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public long lastIndexOf(final Object ok) { return lastIndexOf(((Float)(ok)).floatValue()); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public Float remove(final long index) { return Float.valueOf(removeFloat(index)); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public void push(Float o) { push(o.floatValue()); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public Float pop() { return Float.valueOf(popFloat()); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public Float top() { return Float.valueOf(topFloat()); }
	/** {@inheritDoc}
	 *
	 * @implSpec This implementation delegates to the corresponding type-specific method.
	 * @deprecated Please use the corresponding type-specific method instead.
	 */
	@Deprecated
	@Override
	public Float peek(int i) { return Float.valueOf(peekFloat(i)); }
	@Override
	public String toString() {
	 final StringBuilder s = new StringBuilder();
	 final FloatIterator i = iterator();
	 long n = size64();
	 float k;
	 boolean first = true;
	 s.append("[");
	 while(n-- != 0) {
	  if (first) first = false;
	  else s.append(", ");
	  k = i.nextFloat();
	   s.append(String.valueOf(k));
	 }
	 s.append("]");
	 return s.toString();
	}
	/** A class implementing a sublist view. */
	public static class FloatSubList extends AbstractFloatBigList implements java.io.Serializable {
	 private static final long serialVersionUID = -7046029254386353129L;
	 /** The list this sublist restricts. */
	 protected final FloatBigList l;
	 /** Initial (inclusive) index of this sublist. */
	 protected final long from;
	 /** Final (exclusive) index of this sublist. */
	 protected long to;
	 public FloatSubList(final FloatBigList l, final long from, final long to) {
	  this.l = l;
	  this.from = from;
	  this.to = to;
	 }
	 private boolean assertRange() {
	  assert from <= l.size64();
	  assert to <= l.size64();
	  assert to >= from;
	  return true;
	 }
	 @Override
	 public boolean add(final float k) {
	  l.add(to, k);
	  to++;
	  assert assertRange();
	  return true;
	 }
	 @Override
	 public void add(final long index, final float k) {
	  ensureIndex(index);
	  l.add(from + index, k);
	  to++;
	  assert assertRange();
	 }
	 @Override
	 public boolean addAll(final long index, final Collection<? extends Float> c) {
	  ensureIndex(index);
	  to += c.size();
	  return l.addAll(from + index, c);
	 }
	 @Override
	 public float getFloat(long index) {
	  ensureRestrictedIndex(index);
	  return l.getFloat(from + index);
	 }
	 @Override
	 public float removeFloat(long index) {
	  ensureRestrictedIndex(index);
	  to--;
	  return l.removeFloat(from + index);
	 }
	 @Override
	 public float set(long index, float k) {
	  ensureRestrictedIndex(index);
	  return l.set(from + index, k);
	 }
	 @Override
	 public long size64() { return to - from; }
	 @Override
	 public void getElements(final long from, final float[][] a, final long offset, final long length) {
	  ensureIndex(from);
	  if (from + length > size64()) throw new IndexOutOfBoundsException("End index (" + from + length + ") is greater than list size (" + size64() + ")");
	  l.getElements(this.from + from, a, offset, length);
	 }
	 @Override
	 public void removeElements(final long from, final long to) {
	  ensureIndex(from);
	  ensureIndex(to);
	  l.removeElements(this.from + from, this.from + to);
	  this.to -= (to - from);
	  assert assertRange();
	 }
	 @Override
	 public void addElements(final long index, final float a[][], long offset, long length) {
	  ensureIndex(index);
	  l.addElements(this.from + index, a, offset, length);
	  this.to += length;
	  assert assertRange();
	 }
	 private final class RandomAccessIter extends FloatBigListIterators.AbstractIndexBasedBigListIterator {
	  // We don't set the minPos to be "from" because we need to call our containing class'
	  // add, set, and remove methods with 0 relative to the start of the sublist, not the
	  // start of the original list.
	  // Thus pos is relative to the start of the SubList, not the start of the original list.
	  RandomAccessIter(long pos) {
	   super(0, pos);
	  }
	  @Override
	  protected final float get(long i) { return l.getFloat(from + i); }
	  // Remember, these are calling SUBLIST's methods, meaning 0 is the start of the sublist for these.
	  @Override
	  protected final void add(long i, float k) { FloatSubList.this.add(i, k); }
	  @Override
	  protected final void set(long i, float k) { FloatSubList.this.set(i, k); }
	  @Override
	  protected final void remove(long i) { FloatSubList.this.removeFloat(i); }
	  @Override
	  protected final long getMaxPos() { return to - from; }
	  @Override
	  public void add(float k) {
	   super.add(k);
	   assert assertRange();
	  }
	  @Override
	  public void remove() {
	   super.remove();
	   assert assertRange();
	  }
	 }
	 private class ParentWrappingIter implements FloatBigListIterator {
	  private FloatBigListIterator parent;
	  ParentWrappingIter(FloatBigListIterator parent) {
	   this.parent = parent;
	  }
	  @Override
	  public long nextIndex() { return parent.nextIndex() - from; }
	  @Override
	  public long previousIndex() { return parent.previousIndex() - from; }
	  @Override
	  public boolean hasNext() { return parent.nextIndex() < to; }
	  @Override
	  public boolean hasPrevious() { return parent.previousIndex() >= from; }
	  @Override
	  public float nextFloat() { if (! hasNext()) throw new NoSuchElementException(); return parent.nextFloat(); }
	  @Override
	  public float previousFloat() { if (! hasPrevious()) throw new NoSuchElementException(); return parent.previousFloat(); }
	  @Override
	  public void add(float k) { parent.add(k); }
	  @Override
	  public void set(float k) { parent.set(k); }
	  @Override
	  public void remove() { parent.remove(); }
	  @Override
	  public long back(long n) {
	   if (n < 0) throw new IllegalArgumentException("Argument must be nonnegative: " + n);
	   long currentPos = parent.previousIndex();
	   long parentNewPos = currentPos - n;
	   // Remember, the minimum acceptable previousIndex is not from but (from - 1), since (from - 1)
	   // means this subList is at the beginning of our sub range.
	   // Same reason why previousIndex()'s minimum for the full list is not 0 but -1.
	   if (parentNewPos < (from - 1)) parentNewPos = (from - 1);
	   long toSkip = parentNewPos - currentPos;
	   return parent.back(toSkip);
	  }
	  @Override
	  public long skip(long n) {
	   if (n < 0) throw new IllegalArgumentException("Argument must be nonnegative: " + n);
	   long currentPos = parent.nextIndex();
	   long parentNewPos = currentPos + n;
	   if (parentNewPos > to) parentNewPos = to;
	   long toSkip = parentNewPos - currentPos;
	   return parent.skip(toSkip);
	  }
	 }
	 @Override
	 public FloatBigListIterator listIterator(final long index) {
	  ensureIndex(index);
	  // If this class wasn't public, then RandomAccessIter would live in SUBLISTRandomAccess,
	  // and the switching would be done in sublist(long, long). However, this is a public class
	  // that may have existing implementors, so to get the benefit of RandomAccessIter class for
	  // for existing uses, it has to be done in this class.
	  return l instanceof java.util.RandomAccess ? new RandomAccessIter(index) : new ParentWrappingIter(l.listIterator(index + from));
	 }
	 @Override
	 public FloatSpliterator spliterator() {
	  return l instanceof java.util.RandomAccess ? new IndexBasedSpliterator (l, from, to) : super.spliterator();
	 }
	 @Override
	 public it.unimi.dsi.fastutil.doubles.DoubleSpliterator doubleSpliterator() {
	  if (l instanceof java.util.RandomAccess) {
	   return FloatSpliterators.widen(spliterator());
	  } else {
	   return super.doubleSpliterator();
	  }
	 }
	 @Override
	 public FloatBigList subList(final long from, final long to) {
	  ensureIndex(from);
	  ensureIndex(to);
	  if (from > to) throw new IllegalArgumentException("Start index (" + from + ") is greater than end index (" + to + ")");
	  // Sadly we have to rewrap this, because if there is a sublist of a sublist, and the
	  // subsublist adds, both sublists need to update their "to" value.
	  return new FloatSubList (this, from, to);
	 }
	 @Override
	 public boolean rem(float k) {
	  long index = indexOf(k);
	  if (index == -1) return false;
	  to--;
	  l.removeFloat(from + index);
	  assert assertRange();
	  return true;
	 }
	 @Override
	 public boolean addAll(final long index, final FloatCollection c) {
	  return super.addAll(index, c);
	 }
	 @Override
	 public boolean addAll(final long index, final FloatBigList l) {
	  return super.addAll(index, l);
	 }
	}
	public static class FloatRandomAccessSubList extends FloatSubList implements java.util.RandomAccess {
	 private static final long serialVersionUID = -107070782945191929L;
	 public FloatRandomAccessSubList(final FloatBigList l, final long from, final long to) {
	  super(l, from, to);
	 }
	 @Override
	 public FloatBigList subList(final long from, final long to) {
	  ensureIndex(from);
	  ensureIndex(to);
	  if (from > to) throw new IllegalArgumentException("Start index (" + from + ") is greater than end index (" + to + ")");
	  // Sadly we have to rewrap this, because if there is a sublist of a sublist, and the
	  // subsublist adds, both sublists need to update their "to" value.
	  return new FloatRandomAccessSubList (this, from, to);
	 }
	}
}
