// Copyright 2000-2021 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.

package org.jetbrains.kotlin.idea.actions.internal


import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.ProgressManager.progress
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.intellij.psi.PsiMethod
import com.intellij.psi.PsiModifier
import org.jetbrains.kotlin.descriptors.*
import org.jetbrains.kotlin.idea.KotlinBundle
import org.jetbrains.kotlin.idea.caches.resolve.findModuleDescriptor
import org.jetbrains.kotlin.idea.codeInsight.DescriptorToSourceUtilsIde
import org.jetbrains.kotlin.idea.util.application.isApplicationInternalMode
import org.jetbrains.kotlin.idea.util.application.runReadAction
import org.jetbrains.kotlin.load.java.isFromJava
import org.jetbrains.kotlin.name.FqName
import org.jetbrains.kotlin.psi.KtFile
import org.jetbrains.kotlin.resolve.descriptorUtil.fqNameUnsafe


class SearchNotPropertyCandidatesAction : AnAction() {
    override fun actionPerformed(e: AnActionEvent) {
        val project = e.project!!
        val psiFile = e.getData(CommonDataKeys.PSI_FILE) as? KtFile ?: return

        val desc = psiFile.findModuleDescriptor()
        val result = Messages.showInputDialog(
            KotlinBundle.message("enter.package.fqname"),
            KotlinBundle.message("search.for.not.property.candidates"), Messages.getQuestionIcon())
        val packageDesc = try {
            val fqName = FqName.fromSegments(result!!.split('.'))
            desc.getPackage(fqName)
        } catch (e: Exception) {
            return
        }

        ProgressManager.getInstance().runProcessWithProgressSynchronously(
            {
                runReadAction {
                    ProgressManager.getInstance().progressIndicator!!.isIndeterminate = true
                    processAllDescriptors(packageDesc, project)
                }
            },
            KotlinBundle.message("searching.for.not.property.candidates"),
            true,
            project
        )
    }


    private fun processAllDescriptors(desc: DeclarationDescriptor, project: Project) {
        val processed = mutableSetOf<DeclarationDescriptor>()
        var pFunctions = 0
        val matchedDescriptors = mutableSetOf<FunctionDescriptor>()

        fun recursive(desc: DeclarationDescriptor) {
            if (desc in processed) return
            progress(
                KotlinBundle.message("step.1.collecting.0.1.2", processed.size, pFunctions, matchedDescriptors.size),
                "$desc"
            )
            when (desc) {

                is ModuleDescriptor -> recursive(desc.getPackage(FqName("java")))
                is ClassDescriptor -> desc.unsubstitutedMemberScope.getContributedDescriptors { true }.forEach(::recursive)
                is PackageViewDescriptor -> desc.memberScope.getContributedDescriptors { true }.forEach(::recursive)

                is FunctionDescriptor -> {
                    if (desc.isFromJava) {
                        val name = desc.fqNameUnsafe.shortName().asString()
                        if (name.length > 3 &&
                            ((name.startsWith("get") && desc.valueParameters.isEmpty() && desc.returnType != null) ||
                                    (name.startsWith("set") && desc.valueParameters.size == 1))
                        ) {
                            if (desc in matchedDescriptors) return
                            matchedDescriptors += desc
                        }
                    }
                    pFunctions++
                    return
                }

            }
            processed += desc
        }
        recursive(desc)
        val resultDescriptors = mutableSetOf<FunctionDescriptor>()
        matchedDescriptors.flatMapTo(resultDescriptors) {
            sequenceOf(it, *(it.overriddenDescriptors.toTypedArray())).asIterable()
        }
        println("Found ${resultDescriptors.size} accessors")


        fun PsiMethod.isTrivial(): Boolean {
            val t = this.text
            val s = t.indexOf('{')
            val e = t.lastIndexOf('}')
            return if (s != e && s != -1) t.substring(s, e).lines().size <= 3 else true
        }


        val descriptorToPsiBinding = mutableMapOf<FunctionDescriptor, PsiMethod>()


        var i = 0
        resultDescriptors.forEach { functionDescriptor ->
            progress(KotlinBundle.message("step.2.0.of.1", i++, resultDescriptors.size), "$functionDescriptor")
            val source = DescriptorToSourceUtilsIde.getAllDeclarations(project, functionDescriptor)
                .filterIsInstance<PsiMethod>()
                .firstOrNull() ?: return@forEach
            val abstract = source.modifierList.hasModifierProperty(PsiModifier.ABSTRACT)
            if (!abstract) {
                if (!source.isTrivial()) {
                    descriptorToPsiBinding[functionDescriptor] = source
                }
            }
        }

        val nonTrivial = mutableSetOf<FunctionDescriptor>()
        i = 0
        descriptorToPsiBinding.forEach { (t, u) ->
            progress(KotlinBundle.message("step.3.0.of.1", i++, descriptorToPsiBinding.size), "$t")
            val descriptors = t.overriddenDescriptors
            var impl = false
            descriptors.forEach {
                val source = DescriptorToSourceUtilsIde.getAllDeclarations(project, it).filterIsInstance<PsiMethod>().firstOrNull()
                if (source != null) {
                    if (source.body != null || source.hasModifierProperty(PsiModifier.ABSTRACT))
                        nonTrivial += it
                    impl = true
                }
            }
            if (u.body != null)
                if (!impl)
                    nonTrivial += t
        }
        nonTrivial.forEach(::println)
        println("Non trivial count: ${nonTrivial.size}")
    }

    override fun update(e: AnActionEvent) {
        if (!isApplicationInternalMode()) {
            e.presentation.isVisible = false
            e.presentation.isEnabled = false
        } else {
            e.presentation.isVisible = true
            e.presentation.isEnabled = e.getData(CommonDataKeys.PSI_FILE) is KtFile
        }
    }

}