// Copyright 2000-2021 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.

package org.jetbrains.kotlin.idea.script

import com.intellij.openapi.project.Project
import org.jetbrains.kotlin.config.CompilerSettings
import org.jetbrains.kotlin.idea.compiler.configuration.KotlinCompilerSettings
import org.jetbrains.kotlin.idea.compiler.configuration.KotlinCompilerSettingsListener
import org.jetbrains.kotlin.idea.core.KotlinPluginDisposable
import org.jetbrains.kotlin.idea.core.script.ScriptDefinitionSourceAsContributor
import org.jetbrains.kotlin.idea.core.script.ScriptDefinitionsManager
import org.jetbrains.kotlin.idea.core.script.loadDefinitionsFromTemplatesByPaths
import org.jetbrains.kotlin.idea.util.application.executeOnPooledThread
import org.jetbrains.kotlin.scripting.definitions.ScriptDefinition
import org.jetbrains.kotlin.scripting.definitions.getEnvironment
import java.io.File
import kotlin.io.path.Path
import kotlin.script.experimental.host.ScriptingHostConfiguration
import kotlin.script.experimental.jvm.defaultJvmScriptingHostConfiguration

class ScriptTemplatesFromCompilerSettingsProvider(
    private val project: Project
) : ScriptDefinitionSourceAsContributor {

    init {
        project.messageBus.connect(KotlinPluginDisposable.getInstance(project))
            .subscribe(KotlinCompilerSettingsListener.TOPIC, object : KotlinCompilerSettingsListener {
                override fun <T> settingsChanged(newSettings: T) {
                    if (newSettings !is CompilerSettings) return

                    executeOnPooledThread {
                        if (project.isDefault || project.isDisposed) return@executeOnPooledThread
                        ScriptDefinitionsManager.getInstance(project).reloadDefinitionsBy(this@ScriptTemplatesFromCompilerSettingsProvider)
                    }
                }
            })
    }

    override val definitions: Sequence<ScriptDefinition>
        get() = KotlinCompilerSettings.getInstance(project).settings.let { kotlinSettings ->
            if (kotlinSettings.scriptTemplates.isBlank()) emptySequence()
            else loadDefinitionsFromTemplatesByPaths(
                templateClassNames = kotlinSettings.scriptTemplates.split(',', ' '),
                templateClasspath = kotlinSettings.scriptTemplatesClasspath.split(File.pathSeparator).map(::Path),
                baseHostConfiguration = ScriptingHostConfiguration(defaultJvmScriptingHostConfiguration) {
                    getEnvironment {
                        mapOf(
                            "projectRoot" to (project.basePath ?: project.baseDir.canonicalPath)?.let(::File)
                        )
                    }
                }
            ).asSequence()
        }

    override val id: String = "KotlinCompilerScriptTemplatesSettings"
}

