// Copyright 2000-2021 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.

package org.jetbrains.kotlin.idea.vcs

import com.intellij.BundleBase.replaceMnemonicAmpersand
import com.intellij.CommonBundle
import com.intellij.ide.plugins.PluginManagerCore
import com.intellij.openapi.extensions.PluginId
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.intellij.openapi.ui.Messages.NO
import com.intellij.openapi.ui.Messages.YES
import com.intellij.openapi.util.Key
import com.intellij.openapi.vcs.CheckinProjectPanel
import com.intellij.openapi.vcs.changes.CommitContext
import com.intellij.openapi.vcs.changes.CommitExecutor
import com.intellij.openapi.vcs.checkin.CheckinHandler
import com.intellij.openapi.vcs.checkin.CheckinHandlerFactory
import com.intellij.openapi.vcs.ui.RefreshableOnComponent
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.ui.NonFocusableCheckBox
import com.intellij.util.PairConsumer
import org.jetbrains.kotlin.idea.KotlinJvmBundle
import org.jetbrains.kotlin.psi.NotNullableUserDataProperty
import java.awt.GridLayout
import java.nio.file.Path
import javax.swing.JComponent
import javax.swing.JPanel
import kotlin.io.path.*

private val BUNCH_PLUGIN_ID = PluginId.getId("org.jetbrains.bunch.tool.idea.plugin")

private var Project.bunchFileCheckEnabled: Boolean
        by NotNullableUserDataProperty(
            Key.create("IS_BUNCH_FILE_CHECK_ENABLED_KOTLIN"),
            !PluginManagerCore.isPluginInstalled(BUNCH_PLUGIN_ID)
        )

class BunchFileCheckInHandlerFactory : CheckinHandlerFactory() {
    override fun createHandler(panel: CheckinProjectPanel, commitContext: CommitContext): CheckinHandler {
        return BunchCheckInHandler(panel)
    }

    class BunchCheckInHandler(private val checkInProjectPanel: CheckinProjectPanel) : CheckinHandler() {
        private val project get() = checkInProjectPanel.project

        override fun getBeforeCheckinConfigurationPanel(): RefreshableOnComponent? {
            if (PluginManagerCore.isPluginInstalled(BUNCH_PLUGIN_ID)) return null
            BunchFileUtils.bunchFile(project) ?: return null

            val bunchFilesCheckBox = NonFocusableCheckBox(replaceMnemonicAmpersand(KotlinJvmBundle.message("check.bunch.files")))
            return object : RefreshableOnComponent {
                override fun getComponent(): JComponent {
                    val panel = JPanel(GridLayout(1, 0))
                    panel.add(bunchFilesCheckBox)
                    return panel
                }

                override fun saveState() {
                    project.bunchFileCheckEnabled = bunchFilesCheckBox.isSelected
                }

                override fun restoreState() {
                    bunchFilesCheckBox.isSelected = project.bunchFileCheckEnabled
                }
            }
        }

        override fun beforeCheckin(
            executor: CommitExecutor?,
            additionalDataConsumer: PairConsumer<Any, Any>?
        ): ReturnResult {
            if (!project.bunchFileCheckEnabled) return ReturnResult.COMMIT

            val extensions = BunchFileUtils.bunchExtension(project)?.toSet() ?: return ReturnResult.COMMIT

            val forgottenFiles = HashSet<Path>()
            val commitFiles = checkInProjectPanel.files.mapNotNull { file -> file.toPath().takeIf { it.isRegularFile() } }.toSet()
            for (file in commitFiles) {
                if (file.extension in extensions) continue

                val parent = file.parent ?: continue
                val name = file.name
                for (extension in extensions) {
                    val bunchFile = parent.resolve("$name.$extension")
                    if (bunchFile !in commitFiles && bunchFile.exists()) {
                        forgottenFiles.add(bunchFile)
                    }
                }
            }

            if (forgottenFiles.isEmpty()) return ReturnResult.COMMIT

            val paths = project.basePath?.let(::Path)?.let { projectBasePath ->
                forgottenFiles.map { it.relativeTo(projectBasePath) }
            } ?: forgottenFiles

            var filePaths = paths.map(Path::pathString).sorted()
            if (filePaths.size > 15) {
                filePaths = filePaths.take(15) + "..."
            }

            when (Messages.showYesNoCancelDialog(
                project,
                KotlinJvmBundle.message(
                    "several.bunch.files.haven.t.been.updated.0.do.you.want.to.review.them.before.commit",
                    filePaths.joinToString("\n")
                ),
                KotlinJvmBundle.message("button.text.forgotten.bunch.files"),
                KotlinJvmBundle.message("button.text.review"),
                KotlinJvmBundle.message("button.text.commit"),
                CommonBundle.getCancelButtonText(),
                Messages.getWarningIcon()
            )) {
                YES -> {
                    return ReturnResult.CLOSE_WINDOW
                }
                NO -> return ReturnResult.COMMIT
            }

            return ReturnResult.CANCEL
        }
    }
}

object BunchFileUtils {
    fun bunchFile(project: Project): VirtualFile? {
        val baseDir = project.baseDir ?: return null
        return baseDir.findChild(".bunch")
    }

    fun bunchExtension(project: Project): List<String>? {
        val bunchFile: VirtualFile = bunchFile(project) ?: return null
        val file = bunchFile.toNioPath()
        if (file.notExists()) return null

        val lines = file.readLines().map { it.trim() }.filter { it.isNotEmpty() }
        if (lines.size <= 1) return null

        return lines.drop(1).map { it.split('_').first() }
    }
}
